/*****************************************************************************/
/* galois.c								     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* A simple implementation of Galois field arithmetics - GF(2^32)	     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* (c) 2008, 2021; Pal, A. <apal@szofi.net>				     */
/*****************************************************************************/

#include <stdint.h>
#include "galois.h"

#define		MASK			0xffffffff
#define		HIGHEST_BIT		0x80000000
#define		PRIMITIVE_POLYNOMIAL	0x00400007

__attribute__ ((noinline)) uint32_t galois_2p32_mul(uint32_t a,uint32_t b)
{
 uint32_t	r;
 for ( r=0 ; a ; a>>=1 )
  {	if ( a&1 )
			r^=b;
	if ( b&0x80000000 )
			b=(b<<1)^PRIMITIVE_POLYNOMIAL;
	else
			b=(b<<1);
  }
 return(r);
}

__attribute__ ((noinline)) int galois_2p32_mul_crt(uint32_t *barr,uint32_t b)
{
 int	i;
 for ( i=0; i<32; i++ )
  {	*barr=b;
	if ( b&0x80000000 )
			b=(b<<1)^PRIMITIVE_POLYNOMIAL;
	else
			b=(b<<1);
	barr++;
  }
 return(0);
}

__attribute__ ((noinline)) uint32_t galois_2p32_mul_arr(uint32_t a,uint32_t *barr)
{
 uint32_t	r;
 for ( r=0 ; a ; a>>=1 )
  {	if ( a&1 )	
		r^=*barr;
	barr++;
  }
 return(r);
}


#ifdef	__arm__
__attribute__ ((naked,noinline)) uint32_t galois_2p32_mul_asm(uint32_t a,uint32_t b)
{
 asm 
  (	"	movs	r2, #64	\n"
	"	movs	r3, r0	\n"
	"	lsls	r2, #16 \n"
	"	movs	r0, #0	\n"
	"	adds	r2, #7	\n"
	"1:	cmp	r3, #0	\n"
	"	beq.n	2f	\n"
	"	lsrs	r3, #1	\n"
	"	bcc.n	3f	\n"
	"	eors	r0, r1	\n"
	"3:	lsls	r1, #1	\n"
	"	bcc.n	1b	\n"
	"	eors	r1, r2	\n"
	"	b.n	1b	\n"
	"2:	bx	lr	\n"
  );
}
#endif


 /*	acc[0] ^= in[0] x b;		(x is multiplication over GF(2^32))
	acc[1] ^= in[1] x b;
 */
#ifdef	__arm__

#define		__MUL_1_BIT			\
	"	lsrs	r6, #1		\n"	\
	"	bcc.n	.+0x04		\n"	\
	"	eors	r4, r2		\n"	\
	"	lsrs	r7, #1		\n"	\
	"	bcc.n	.+0x04		\n"	\
	"	eors	r5, r2		\n"	\
	"	lsls	r2, #1		\n"	\
	"	bcc.n	.+0x04		\n"	\
	"	eors	r2, r3		\n"

#define		__MUL_2_BIT	__MUL_1_BIT	__MUL_1_BIT
#define		__MUL_4_BIT	__MUL_2_BIT	__MUL_2_BIT

__attribute__ ((naked,noinline)) void galois_2p32_mul_accumulate_2(uint32_t *acc,uint32_t *in,uint32_t b)
{
 asm 
  (	"	push	{r4-r7}		\n"
	"	movs	r3, #64		\n"
	"	ldmia	r0!, {r4, r5}	\n"
	"	lsls	r3, #16 	\n"
	"	ldmia	r1!, {r6, r7}	\n"
	"	movs	r1, #32		\n"
	"	adds	r3, #7		\n"
	"	subs	r0, #8		\n"
	"0:				\n"
		__MUL_4_BIT
	"	subs	r1, #4		\n"
	"	bne.n	0b		\n"
	"	stmia	r0!, {r4, r5}	\n"
	"	pop	{r4-r7}		\n"
	"	bx	lr		\n"

  );
}

#endif


__attribute__ ((noinline)) uint32_t galois_2p32_pow(uint32_t a,int nn) 
{
 uint32_t	r,m;
 unsigned	n;
 if ( nn==0 )
	return(1);
 else if ( a==0 )
	return(0);
 for ( r=1,m=a,n=(unsigned)nn ; n ; n>>=1 )
  {	
#ifdef	__arm__
	if ( n&1 )	r=galois_2p32_mul_asm(r,m);
	m=galois_2p32_mul_asm(m,m);
#else
	if ( n&1 )	r=galois_2p32_mul(r,m);
	m=galois_2p32_mul(m,m);
#endif

  }
 return(r);
}

uint32_t galois_2p32_div(uint32_t a,uint32_t b)
{
 b=galois_2p32_pow(b,MASK-1);
 return(galois_2p32_mul(a,b));
}

uint32_t galois_2p32_rcp(uint32_t b)
{
 return(galois_2p32_pow(b,MASK-1));
}

