/*****************************************************************************/
/* gmath.c 								     */
/*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>

#include "galois.h"
#include "gmath.h"

/*****************************************************************************/

void *matrix_alloc_gen(int nrow,int ncol,size_t msize)
{
 int	i;

 void **m;
 m=(void **)malloc(sizeof(void *)*(nrow+1));
 memset(m,0,sizeof(sizeof(void *)*(nrow+1)));
 if ( m==NULL )	return(NULL);
 for ( i=0 ; i<nrow ; i++ )
  {	m[i]=(void *)malloc(msize*ncol);
	if ( m[i]==NULL )
	 {	matrix_free(m);
		return(NULL);
	 }
  }
 m[nrow]=NULL;
 return((void *)m);
}

void *matrix_alloc(int d,size_t msize)
{
 void	*m;
 m=matrix_alloc_gen(d,d,msize);
 return(m);
}

void matrix_free(void *vm)
{
 void	**m=(void **)vm;
 int	i;
 for ( i=0 ; m[i] != NULL ; i++ )
  {	free(m[i]);		}
 free(m);
}
void *vector_alloc(int d,size_t msize)
{
 void *v;
 v=(double *)malloc(msize*d);
 return(v);
}
void vector_free(void *v)
{
 free(v);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void galois_g2p32_swap_matrix_rows(uint32_t **a,int r,int s)
{
 uint32_t	*t;
 if ( r==s )	return;
 t=a[r],a[r]=a[s],a[s]=t;
}
static void galois_g2p32_swap_vector_elements(uint32_t *v,int r,int s)
{
 uint32_t	t;
 if ( r==s )	return;
 t=v[r],v[r]=v[s],v[s]=t;
}

int galois_2p32_solve_gauss(uint32_t **a,uint32_t *b,int n)
{
 int	i;
 
 for ( i=0 ; i<n ; i++ )
  {	int	j;

	for ( j=i ; j<n ; j++ )
	 {	if ( a[j][i] != 0 )	
			break;		
	 }
	if ( j==n )
		return(-1);

	galois_g2p32_swap_matrix_rows(a,i,j);
	galois_g2p32_swap_vector_elements(b,i,j);

	for ( j=i+1 ; j<n ; j++ )
	 {	int		k;
		uint32_t	t;

		t=galois_2p32_div(a[j][i],a[i][i]);

		for ( k=0 ; k<n ; k++ )
		 {	a[j][k] ^= galois_2p32_mul(a[i][k],t);
	  	 }
		b[j] ^= galois_2p32_mul(b[i],t);
	 }
  }
 for ( i=n-1 ; i>=0 ; i-- )
  {	int		j;
	uint32_t	t;
	t=b[i];
	for ( j=i+1 ; j<n ; j++ )
	 {	t ^= galois_2p32_mul(a[i][j],b[j]);		
	 }
	b[i]=galois_2p32_div(t,a[i][i]);
  }
 return(0);
}

int galois_2p32_invert_gauss(uint32_t **a,int n)
{
 int		i;
 uint32_t	**e;

 e=matrix_alloc(n,sizeof(uint32_t));

 for ( i=0 ; i<n ; i++ )
  {	int	j;
	for ( j=0 ; j<n ; j++ )
	 {	e[i][j]=0;		}
	e[i][i]=1;
  }

 for ( i=0 ; i<n ; i++ )
  {	int	j;
	uint32_t	iaii;

	for ( j=i ; j<n ; j++ )
	 {	if ( a[j][i] != 0 )	
			break;		
	 }
	if ( j==n )	
		return(-1);

	galois_g2p32_swap_matrix_rows(a,i,j);
	galois_g2p32_swap_matrix_rows(e,i,j);

	iaii=galois_2p32_rcp(a[i][i]);

	for ( j=i+1 ; j<n ; j++ )
	 {	uint32_t	t;
		int		k;

		t=galois_2p32_mul(a[j][i],iaii);
		for ( k=0 ; k<n ; k++ )
		 {	a[j][k] ^= galois_2p32_mul(t,a[i][k]);
			e[j][k] ^= galois_2p32_mul(t,e[i][k]);
		 }
	 }
  }
 for ( i=n-1 ; i>=0 ; i-- )
  {	int		k;
	uint32_t	iaii;

	iaii=galois_2p32_rcp(a[i][i]);

	for ( k=0 ; k<n ; k++ )
	 {	uint32_t	t;
		int		j;

		t=e[i][k];
		for ( j=i+1 ; j<n ; j++ )
		 {	t ^= galois_2p32_mul(e[j][k],a[i][j]);
		 }
		e[i][k]=galois_2p32_mul(t,iaii);
	 }
  }

 for ( i=0 ; i<n ; i++ )
  {	int	j;
	for ( j=0 ; j<n ; j++ )
	 {	a[i][j]=e[i][j];		}
  }

 matrix_free(e);
 
 return(0);
}

/*****************************************************************************/

static void swap_columns(uint32_t **m,int nrow,int c1,int c2)
{
 while ( 0<nrow )
  {	uint32_t	t,*row;
	row=*m++;
	t=row[c1];
	row[c1]=row[c2];
	row[c2]=t;
	nrow--;
  }
}

int galois_2p32_column_echelon(uint32_t **m,int nrow,int ncol)
{
 int	c,lead;

 lead=0;
 for ( c=0; c<ncol; c++ )
  {	int		i,j,k;
	uint32_t	t;
	if ( nrow <= lead )
	 {	return(0);
	 }
	i=c;
	while ( m[lead][i]==0 )
	 {	i++;
		if ( ncol==i )
		 {	i=c;
			lead++;
			if ( nrow==lead )
			 {	return(0);
			 }
		 }
	 }
	if ( i != c )
	 {	swap_columns(m,nrow,i,c);

	 }
	t=galois_2p32_rcp(m[lead][c]);
	for ( k=0; k<nrow; k++ )
	 {	m[k][c]=galois_2p32_mul(m[k][c],t);
	 }
	for ( j=0; j<ncol; j++ )
	 {	uint32_t	t;
		if ( j==c )
			continue;
		t=m[lead][j];
		for ( k=0; k<nrow; k++ )
		 {	m[k][j] ^= galois_2p32_mul(m[k][c],t);
		 }
	 }
	lead++;
  }

 return(0);
} 

int galois_2p32_row_echelon(uint32_t **m,int nrow,int ncol)
{
 int	c,lead;

 lead=0;
 for ( c=0; c<nrow; c++ )
  {	int		i,j,k;
	uint32_t	t;
	if ( ncol <= lead )
	 {	return(0);
	 }
	i=c;
	while ( m[i][lead]==0 )
	 {	i++;
		if ( nrow==i )
		 {	i=c;
			lead++;
			if ( ncol==lead )
			 {	return(0);
			 }
		 }
	 }
	if ( i != c )
	 {	uint32_t	*t;
		t=m[i];m[i]=m[c];m[c]=t;
	 }
	t=galois_2p32_rcp(m[c][lead]);
	for ( k=0; k<ncol; k++ )
	 {	m[c][k]=galois_2p32_mul(m[c][k],t);
	 }
	for ( j=0; j<nrow; j++ )
	 {	uint32_t	t;
		if ( j==c )
			continue;
		t=m[j][lead];
		for ( k=0; k<ncol; k++ )
		 {	m[j][k] ^= galois_2p32_mul(m[c][k],t);
		 }
	 }
	lead++;
  }

 return(0);
} 

/*****************************************************************************/
